# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

import jax.numpy as jnp
from jax import grad, jit

from pyhgf.typing import Edges


@partial(jit, static_argnames=("edges", "node_idx"))
def posterior_update_mean_continuous_node(
    attributes: Dict,
    edges: Edges,
    node_idx: int,
    node_precision: float,
) -> float:
    r"""Update the mean of a state node using the value prediction errors.

    1. Mean update from value coupling.

    The new mean of a state node :math:`b` value coupled with other input and/or state
    nodes :math:`j` at time :math:`k` is given by:

    For linear value coupling:

    .. math::
        \mu_b^{(k)} =  \hat{\mu}_b^{(k)} + \sum_{j=1}^{N_{children}}
            \frac{\kappa_j \hat{\pi}_j^{(k)}}{\pi_b} \delta_j^{(k)}

    Where :math:`\kappa_j` is the volatility coupling strength between the child node
    and the state node and :math:`\delta_j^{(k)}` is the value prediction error that
    was computed beforehand by
    :py:func:`pyhgf.updates.prediction_errors.continuous.continuous_node_value_prediction_error`.

    For non-linear value coupling:

    .. math::
        \mu_b^{(k)} =  \hat{\mu}_b^{(k)} + \sum_{j=1}^{N_{children}}
            \frac{\kappa_j g'_{j,b}({\mu}_b^{(k-1)}) \hat{\pi}_j^{(k)}}{\pi_b}
            \delta_j^{(k)}


    2. Mean update from volatility coupling.

    The new mean of a state node :math:`b` volatility coupled with other input and/or
    state nodes :math:`j` at time :math:`k` is given by:

    .. math::
        \mu_b^{(k)} = \hat{\mu}_b^{(k)} + \frac{1}{2\pi_b}
          \sum_{j=1}^{N_{children}} \kappa_j \gamma_j^{(k)} \Delta_j^{(k)}

    where :math:`\kappa_j` is the volatility coupling strength between the volatility
    parent and the volatility children :math:`j` and :math:`\Delta_j^{(k)}` is the
    volatility prediction error is given by:

    .. math::

        \Delta_j^{(k)} = \frac{\hat{\pi}_j^{(k)}}{\pi_j^{(k)}} +
        \hat{\pi}_j^{(k)} \left( \delta_j^{(k)} \right)^2 - 1

    with :math:`\delta_j^{(k)}` the value prediction error
    :math:`\delta_j^{(k)} = \mu_j^{k} - \hat{\mu}_j^{k}`.

    :math:`\gamma_j^{(k)}` is the effective precision of the prediction, given by:

    .. math::

        \gamma_j^{(k)} = \Omega_j^{(k)} \hat{\pi}_j^{(k)}

    with :math:`\Omega_j^{(k)}` the predicted volatility computed in the prediction
    step :py:func:`pyhgf.updates.prediction.predict_precision`.

    If the child node is a continuous state node, the volatility prediction error
    :math:`\Delta_j^{(k)}` was computed by
    :py:func:`pyhgf.updates.prediction_errors.continuous.continuous_node_volatility_prediction_error`.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.
    node_idx :
        Pointer to the value parent node that will be updated.
    node_precision :
        The precision of the node. Depending on the kind of volatility update, this
        value can be the expected precision (ehgf), or the posterior from the update
        (standard).

    Returns
    -------
    posterior_mean :
        The new posterior mean.

    Notes
    -----
    This update step is similar to the one used for the state node, except that it uses
    the observed value instead of the mean of the child node, and the expected mean of
    the parent node instead of the expected mean of the child node.

    See Also
    --------
    posterior_update_precision_continuous_node

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # sum the prediction errors from both value and volatility coupling
    (
        value_precision_weigthed_prediction_error,
        volatility_precision_weigthed_prediction_error,
    ) = (0.0, 0.0)

    # Value coupling updates - update the mean of a value parent
    # ----------------------------------------------------------
    if edges[node_idx].value_children is not None:
        for value_child_idx, value_coupling, coupling_fn in zip(
            edges[node_idx].value_children,  # type: ignore
            attributes[node_idx]["value_coupling_children"],
            edges[node_idx].coupling_fn,
        ):
            # get the value prediction error (VAPE)
            # if this is jnp.nan (no observation) set the VAPE to 0.0
            value_prediction_error = attributes[value_child_idx]["temp"][
                "value_prediction_error"
            ]

            # cancel the prediction error if the child value was not observed
            value_prediction_error *= attributes[value_child_idx]["observed"]

            # get differential of coupling function with value children
            if coupling_fn is None:  # linear coupling
                coupling_fn_prime = 1
            else:  # non-linear coupling
                # Compute the derivative of the coupling function
                coupling_fn_prime = grad(coupling_fn)(attributes[node_idx]["mean"])

            # expected precisions from the value children
            # sum the precision weigthed prediction errors over all children
            value_precision_weigthed_prediction_error += (
                (
                    (
                        value_coupling
                        * attributes[value_child_idx]["expected_precision"]
                        * coupling_fn_prime
                    )
                    / node_precision
                )
            ) * value_prediction_error

    # Volatility coupling updates - update the mean of a volatility parent
    # --------------------------------------------------------------------
    if edges[node_idx].volatility_children is not None:
        for volatility_child_idx, volatility_coupling in zip(
            edges[node_idx].volatility_children,  # type: ignore
            attributes[node_idx]["volatility_coupling_children"],
        ):
            # get the volatility prediction error (VOPE)
            volatility_prediction_error = attributes[volatility_child_idx]["temp"][
                "volatility_prediction_error"
            ]

            # retrieve the effective precision (γ)
            # computed during the prediction step
            effective_precision = attributes[volatility_child_idx]["temp"][
                "effective_precision"
            ]

            # the precision weigthed prediction error
            precision_weigthed_prediction_error = (
                volatility_coupling * effective_precision * volatility_prediction_error
            )

            # weight using the node's precision
            precision_weigthed_prediction_error *= 1 / (2 * node_precision)

            # cancel the prediction error if the child value was not observed
            precision_weigthed_prediction_error *= attributes[volatility_child_idx][
                "observed"
            ]

            # aggregate over volatility children
            volatility_precision_weigthed_prediction_error += (
                precision_weigthed_prediction_error
            )

    # Compute the new posterior mean
    # using value prediction errors from both value and volatility coupling
    posterior_mean = (
        attributes[node_idx]["expected_mean"]
        + value_precision_weigthed_prediction_error
        + volatility_precision_weigthed_prediction_error
    )

    return posterior_mean


@partial(jit, static_argnames=("edges", "node_idx"))
def posterior_update_precision_continuous_node(
    attributes: Dict,
    edges: Edges,
    node_idx: int,
) -> float:
    r"""Update the precision of a state node using the volatility prediction errors.

    #. Precision update from value coupling.

    The new precision of a state node :math:`b` value coupled with other input and/or
    state nodes :math:`j` at time :math:`k` is given by:

    For linear coupling (default)

    .. math::

            \pi_b^{(k)} = \hat{\pi}_b^{(k)} + \sum_{j=1}^{N_{children}}
            \kappa_j^2 \hat{\pi}_j^{(k)}

    Where :math:`\kappa_j` is the volatility coupling strength between the child node
    and the state node and :math:`\delta_j^{(k)}` is the value prediction error that
    was computed before hand by
    :py:func:`pyhgf.updates.prediction_errors.continuous.continuous_node_value_prediction_error`.

    For non-linear value coupling:

    .. math::

            \pi_b^{(k)} = \hat{\pi}_b^{(k)} + \sum_{j=1}^{N_{children}}
            \hat{\pi}_j^{(k)} * (\kappa_j^2 * g'_{j,b}(\mu_b^(k-1))^2 -
            g''_{j,b}(\mu_b^(k-1))*\delta_j)

    #. Precision update from volatility coupling.

    The new precision of a state node :math:`b` volatility coupled with other input
    and/or state nodes :math:`j` at time :math:`k` is given by:

    .. math::

        \pi_b^{(k)} = \hat{\pi}_b^{(k)} + \sum_{j=1}^{N_{children}}
        \frac{1}{2} \left( \kappa_j \gamma_j^{(k)} \right) ^2 +
        \left( \kappa_j \gamma_j^{(k)} \right) ^2 \Delta_j^{(k)} -
        \frac{1}{2} \kappa_j^2 \gamma_j^{(k)} \Delta_j^{(k)}

    where :math:`\kappa_j` is the volatility coupling strength between the volatility
    parent and the volatility children :math:`j` and :math:`\Delta_j^{(k)}` is the
    volatility prediction error given by:

    .. math::

        \Delta_j^{(k)} = \frac{\hat{\pi}_j^{(k)}}{\pi_j^{(k)}} +
        \hat{\pi}_j^{(k)} \left( \delta_j^{(k)} \right)^2 - 1

    with :math:`\delta_j^{(k)}` the value prediction error
    :math:`\delta_j^{(k)} = \mu_j^{k} - \hat{\mu}_j^{k}`.

    :math:`\gamma_j^{(k)}` is the effective precision of the prediction, given by:

    .. math::

        \gamma_j^{(k)} = \Omega_j^{(k)} \hat{\pi}_j^{(k)}

    that was computed in the prediction step.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index list value and volatility parents and children.
    node_idx :
        Pointer to the value parent node that will be updated.
    time_step :
        The time elapsed between this observation and the previous one.

    Returns
    -------
    posterior_precision :
        The new posterior precision.

    Notes
    -----
    This update step is similar to the one used for the state node, except that it uses
    the observed value instead of the mean of the child node, and the expected mean of
    the parent node instead of the expected mean of the child node.

    See Also
    --------
    posterior_update_mean_continuous_node

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # sum the prediction errors from both value and volatility coupling
    precision_weigthed_prediction_error = 0.0

    # Value coupling updates - update the precision of a value parent
    # ---------------------------------------------------------------
    if edges[node_idx].value_children is not None:
        for value_child_idx, value_coupling, coupling_fn in zip(
            edges[node_idx].value_children,  # type: ignore
            attributes[node_idx]["value_coupling_children"],
            edges[node_idx].coupling_fn,
        ):
            if coupling_fn is None:  # linear coupling
                coupling_fn_prime = 1
                coupling_fn_second = 0
            else:  # non-linear coupling
                coupling_fn_prime = grad(coupling_fn)(attributes[node_idx]["mean"]) ** 2
                value_prediction_error = attributes[value_child_idx]["temp"][
                    "value_prediction_error"
                ]
                coupling_fn_second = (
                    grad(grad(coupling_fn))(attributes[node_idx]["mean"])
                    * value_prediction_error
                )

            # cancel the prediction error if the child value was not observed
            precision_weigthed_prediction_error += (
                value_coupling**2
                * attributes[value_child_idx]["expected_precision"]
                * coupling_fn_prime
                - coupling_fn_second
            ) * attributes[value_child_idx]["observed"]

    # Volatility coupling updates - update the precision of a volatility parent
    # -------------------------------------------------------------------------
    if edges[node_idx].volatility_children is not None:
        for volatility_child_idx, volatility_coupling in zip(
            edges[node_idx].volatility_children,  # type: ignore
            attributes[node_idx]["volatility_coupling_children"],
        ):
            # volatility weigthed precision for the volatility child (γ)
            effective_precision = attributes[volatility_child_idx]["temp"][
                "effective_precision"
            ]

            # retrieve the volatility prediction error
            volatility_prediction_error = attributes[volatility_child_idx]["temp"][
                "volatility_prediction_error"
            ]

            # sum over all volatility children
            # cancel the prediction error if the child value was not observed
            precision_weigthed_prediction_error += (
                0.5 * (volatility_coupling * effective_precision) ** 2
                + (volatility_coupling * effective_precision) ** 2
                * volatility_prediction_error
                - 0.5
                * volatility_coupling**2
                * effective_precision
                * volatility_prediction_error
            ) * attributes[volatility_child_idx]["observed"]

    # Compute the new posterior precision
    # using value prediction errors from both value and volatility coupling
    posterior_precision = (
        attributes[node_idx]["expected_precision"] + precision_weigthed_prediction_error
    )

    # ensure the new precision is greater than 0
    observed_posterior_precision = jnp.where(
        posterior_precision > 1e-128, posterior_precision, jnp.nan
    )

    # additionnal steps for unobserved values
    # ---------------------------------------

    # List the node's volatility parents
    volatility_parents_idxs = edges[node_idx].volatility_parents

    # Get the tonic volatility from the node
    total_volatility = attributes[node_idx]["tonic_volatility"]

    # Look at the (optional) volatility parents and add their value to the tonic
    # volatility to get the total volatility
    if volatility_parents_idxs is not None:
        for volatility_parents_idx, volatility_coupling in zip(
            volatility_parents_idxs,
            attributes[node_idx]["volatility_coupling_parents"],
        ):
            total_volatility += (
                volatility_coupling * attributes[volatility_parents_idx]["mean"]
            )

    # compute the predicted_volatility from the total volatility
    time_step = attributes[-1]["time_step"]
    predicted_volatility = time_step * jnp.exp(total_volatility)

    # Estimate the new precision for the continuous state node
    unobserved_posterior_precision = 1 / (
        (1 / attributes[node_idx]["precision"]) + predicted_volatility
    )

    # for all children, look at the values of VAPE
    # if all these values are NaNs, the node has not received observations
    observations = []
    if edges[node_idx].value_children is not None:
        for children_idx in edges[node_idx].value_children:  # type: ignore
            observations.append(attributes[children_idx]["observed"])
    if edges[node_idx].volatility_children is not None:
        for children_idx in edges[node_idx].volatility_children:  # type: ignore
            observations.append(attributes[children_idx]["observed"])
    observations = jnp.any(jnp.array(observations))

    posterior_precision = (
        unobserved_posterior_precision * (1 - observations)  # type: ignore
        + observed_posterior_precision * observations
    )

    return posterior_precision


@partial(jit, static_argnames=("edges", "node_idx"))
def continuous_node_posterior_update(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Update the posterior of a continuous node using the standard HGF update.

    The standard HGF posterior update is a two-step process:
    1. Update the posterior precision.
    2. Update the posterior mean and assume that the posterior precision is the value
    updated in the first step.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the node that needs to be updated. After continuous updates, the
        parameters of value and volatility parents (if any) will be different.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index list value and volatility parents and children.

    Returns
    -------
    attributes :
        The updated attributes of the probabilistic nodes.

    See Also
    --------
    posterior_update_precision_continuous_node, posterior_update_mean_continuous_node

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # update the posterior mean and precision
    attributes[node_idx]["precision"] = posterior_update_precision_continuous_node(
        attributes, edges, node_idx
    )

    attributes[node_idx]["mean"] = posterior_update_mean_continuous_node(
        attributes, edges, node_idx, node_precision=attributes[node_idx]["precision"]
    )

    return attributes


@partial(jit, static_argnames=("edges", "node_idx"))
def continuous_node_posterior_update_ehgf(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Update the posterior of a continuous node using the eHGF update.

    The eHGF posterior update is a two-step process:
    1. Update the posterior mean and assume that the posterior precision is equal to
    the expected precision.
    2. Update the posterior precision.

    .. note::
        By updating the mean first, and approximating the precision using the expected,
        precision, this update step often perform better than the regular update and
        limit the occurence of negative precision that cause the model to fail under
        some circumstances

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the node that needs to be updated. After continuous updates, the
        parameters of value and volatility parents (if any) will be different.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index list value and volatility parents and children.

    Returns
    -------
    attributes :
        The updated attributes of the probabilistic nodes.

    See Also
    --------
    posterior_update_precision_continuous_node, posterior_update_mean_continuous_node

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # update the posterior mean and precision using the eHGF update step
    # we start with the mean update using the expected precision as an approximation
    posterior_mean = posterior_update_mean_continuous_node(
        attributes,
        edges,
        node_idx,
        node_precision=attributes[node_idx]["expected_precision"],
    )
    attributes[node_idx]["mean"] = posterior_mean

    posterior_precision = posterior_update_precision_continuous_node(
        attributes,
        edges,
        node_idx,
    )
    attributes[node_idx]["precision"] = posterior_precision

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes

if TYPE_CHECKING:
    from graphviz.sources import Source

    from pyhgf.model import Network


def plot_trajectories(
    network: "Network",
    ci: bool = True,
    show_surprise: bool = True,
    show_posterior: bool = False,
    show_total_surprise: bool = False,
    figsize: Tuple[int, int] = (18, 9),
    axs: Optional[Union[List, Axes]] = None,
) -> Axes:
    r"""Plot the trajectories of the nodes' sufficient statistics and surprise.

    This function will plot the expected mean and precision (converted into standard
    deviation) and the surprise at each level of the node structure.

    Parameters
    ----------
    network :
        An instance of the main Network class.
    ci :
        Show the uncertainty around the values estimates (standard deviation).
    show_surprise :
        If `True` plot each node's surprise together with sufficient statistics.
        If `False`, only the input node's surprise is depicted.
    show_posterior :
        If `True`, plot the posterior mean and precision on the top of expected mean and
        precision. Defaults to `False`.
    show_total_surprise :
        If `True`, plot the sum of surprises across all nodes in the bottom panel.
        Defaults to `False`.
    figsize :
        The width and height of the figure. Defaults to `(18, 9)` for a two-level model,
        or to `(18, 12)` for a three-level model.
    axs :
        A list of Matplotlib axes instances where to draw the trajectories. This should
        correspond to the number of nodes in the structure. The default is `None`
        (create a new figure).

    Returns
    -------
    axs :
        The Matplotlib axes instances where to plot the trajectories.

    Examples
    --------
    Visualization of nodes' trajectories from a three-level continuous HGF model.

    .. plot::

        from pyhgf import load_data
        from pyhgf.model import HGF

        # Set up standard 3-level HGF for continuous inputs
        hgf = HGF(
            n_levels=3,
            model_type="continuous",
            initial_mean={"1": 1.04, "2": 1.0, "3": 1.0},
            initial_precision={"1": 1e4, "2": 1e1, "3": 1e1},
            tonic_volatility={"1": -13.0, "2": -2.0, "3": -2.0},
            tonic_drift={"1": 0.0, "2": 0.0, "3": 0.0},
            volatility_coupling={"1": 1.0, "2": 1.0},
        )

        # Read USD-CHF data
        timeserie = load_data("continuous")

        # Feed input
        hgf.input_data(input_data=timeserie)

        # Plot
        hgf.plot_trajectories();

    Visualization of nodes' trajectories from a three-level binary HGF model.

    .. plot::

        from pyhgf import load_data
        from pyhgf.model import HGF
        import jax.numpy as jnp

        # Read binary input
        u, _ = load_data("binary")

        three_levels_hgf = HGF(
            n_levels=3,
            model_type="binary",
            initial_mean={"1": .0, "2": .5, "3": 0.},
            initial_precision={"1": .0, "2": 1e4, "3": 1e1},
            tonic_volatility={"1": None, "2": -6.0, "3": -2.0},
            tonic_drift={"1": None, "2": 0.0, "3": 0.0},
            volatility_coupling={"1": None, "2": 1.0},
            eta0=0.0,
            eta1=1.0,
            binary_precision = jnp.inf,
        )

        # Feed input
        three_levels_hgf = three_levels_hgf.input_data(u)

        # Plot
        three_levels_hgf.plot_trajectories();

    """
    trajectories_df = network.to_pandas()
    n_nodes = len(network.edges)

    if axs is None:
        _, axs = plt.subplots(
            nrows=n_nodes + 1 if show_total_surprise else n_nodes,
            figsize=figsize,
            sharex=True,
        )

    # plot all nodes
    # --------------
    ax_i = n_nodes - 1
    for node_idx in range(n_nodes):

        if node_idx in network.input_idxs:
            _show_posterior = True
            color = "#4c72b0"
        else:
            _show_posterior = show_posterior
            color = (
                "#55a868"
                if network.edges[node_idx].volatility_children is None
                else "#c44e52"
            )

        # use different colors for each node
        plot_nodes(
            network=network,
            node_idxs=node_idx,
            axs=axs[ax_i],
            color=color,
            show_surprise=show_surprise,
            show_posterior=_show_posterior,
            ci=ci,
        )
        ax_i -= 1

    # plot the total surprise of the model
    # ------------------------------------
    if show_total_surprise:
        surprise_ax = axs[n_nodes].twinx()
        surprise_ax.fill_between(
            x=trajectories_df.time,
            y1=trajectories_df.total_surprise,
            y2=trajectories_df.total_surprise.min(),
            label="Surprise",
            color="#7f7f7f",
            alpha=0.2,
        )
        surprise_ax.plot(
            trajectories_df.time,
            trajectories_df.total_surprise,
            color="#2a2a2a",
            linewidth=0.5,
            zorder=-1,
            label="Surprise",
        )
        sp = trajectories_df.total_surprise.sum()
        surprise_ax.set_title(f"Total surprise: {sp:.2f}", loc="right")
        surprise_ax.set_ylabel("Surprise")

    axs[n_nodes - 1].set_xlabel("Time")

    return axs


def plot_correlations(network: "Network") -> Axes:
    """Plot the heatmap correlation of the sufficient statistics trajectories.

    Parameters
    ----------
    network :
        An instance of the HGF model.

    Returns
    -------
    axs :
        The Matplotlib axe instance containing the heatmap of parameters trajectories
        correlation.

    """
    trajectories_df = network.to_pandas()
    trajectories_df = pd.concat(
        [
            trajectories_df[["time"]],
            trajectories_df[
                [
                    f"x_{i}_mean"
                    for i in range(len(network.edges))
                    if i in network.input_idxs
                ]
            ],
            trajectories_df.filter(regex="expected"),
            trajectories_df.filter(regex="surprise"),
        ],
        axis=1,
    )

    correlation_mat = trajectories_df.corr()
    ax = sns.heatmap(
        correlation_mat,
        cmap="RdBu",
        vmin=-1,
        vmax=1,
        linewidths=2,
        square=True,
    )
    ax.set_title("Correlations between the model trajectories")
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha="right", size=8)
    ax.set_yticklabels(ax.get_yticklabels(), size=8)

    return ax


def plot_network(network: "Network") -> "Source":
    """Visualization of node network using GraphViz.

    Parameters
    ----------
    network :
        An instance of main Network class.

    Notes
    -----
    This function requires [Graphviz](https://github.com/xflr6/graphviz) to be
    installed to work correctly.

    """
    try:
        import graphviz
    except ImportError:
        print(
            (
                "Graphviz is required to plot networks. "
                "See https://pypi.org/project/graphviz/"
            )
        )

    graphviz_structure = graphviz.Digraph("hgf-nodes", comment="Nodes structure")

    graphviz_structure.attr("node", shape="circle")

    # create the rest of nodes
    for idx in range(len(network.edges)):

        style = "filled" if idx in network.input_idxs else ""

        if network.edges[idx].node_type == 0:
            # binary state node
            graphviz_structure.node(
                f"x_{idx}", label=str(idx), shape="ellipse", style=style
            )

        elif network.edges[idx].node_type == 1:
            # binary state node
            graphviz_structure.node(
                f"x_{idx}", label=str(idx), shape="square", style=style
            )

        elif network.edges[idx].node_type == 2:
            # Continuous state nore
            graphviz_structure.node(
                f"x_{idx}", label=str(idx), shape="circle", style=style
            )

        elif network.edges[idx].node_type == 3:
            # Exponential family state nore
            graphviz_structure.node(
                f"x_{idx}",
                label=f"EF-{idx}",
                style="filled",
                shape="circle",
                fillcolor="#ced6e4",
            )

        elif network.edges[idx].node_type == 4:
            # Dirichlet Process state node
            graphviz_structure.node(
                f"x_{idx}",
                label=f"DP-{idx}",
                style="filled",
                shape="doublecircle",
                fillcolor="#e2d8c1",
            )

        elif network.edges[idx].node_type == 5:
            # Categorical state node
            graphviz_structure.node(
                f"x_{idx}",
                label=f"Ca-{idx}",
                style=style,
                shape="diamond",
                fillcolor="#e2d8c1",
            )

    # connect value parents
    for i, index in enumerate(network.edges):
        value_parents = index.value_parents

        if value_parents is not None:
            for value_parents_idx in value_parents:

                # get the coupling function from the value parent
                child_idx = network.edges[value_parents_idx].value_children.index(i)
                coupling_fn = network.edges[value_parents_idx].coupling_fn[child_idx]
                graphviz_structure.edge(
                    f"x_{value_parents_idx}",
                    f"x_{i}",
                    color="black" if coupling_fn is None else "black:invis:black",
                )

    # connect volatility parents
    for i, index in enumerate(network.edges):
        volatility_parents = index.volatility_parents

        if volatility_parents is not None:
            for volatility_parents_idx in volatility_parents:
                graphviz_structure.edge(
                    f"x_{volatility_parents_idx}",
                    f"x_{i}",
                    color="gray",
                    style="dashed",
                    arrowhead="dot",
                )

    # unflat the structure to better handle large/uneven networks
    graphviz_structure = graphviz_structure.unflatten(stagger=3)

    return graphviz_structure


def plot_nodes(
    network: "Network",
    node_idxs: Union[int, List[int]],
    ci: bool = True,
    show_surprise: bool = True,
    show_posterior: bool = False,
    figsize: Tuple[int, int] = (12, 5),
    color: Optional[Union[Tuple, str]] = None,
    axs: Optional[Union[List, Axes]] = None,
):
    r"""Plot the trajectory of expected sufficient statistics of a set of nodes.

    This function will plot the expected mean and precision (converted into standard
    deviation) before observation, and the Gaussian surprise after observation. If
    `children_inputs` is `True`, will also plot the children input (mean for value
    coupling and precision for volatility coupling).

    Parameters
    ----------
    network :
        An instance of main Network class.
    node_idxs :
        The index(es) of the probabilistic node(s) that should be plotted. If multiple
        indexes are provided, multiple rows will be appended to the figure, one for
        each node.
    ci :
        Whether to show the uncertainty around the values estimates (using the standard
        deviation :math:`\sqrt{\frac{1}{\hat{\pi}}}`).
    show_surprise :
        If `True` the surprise, defined as the negative log probability of the
        observation given the expectation, is plotted in the backgroud of the figure
        as grey shadded area.
    show_posterior :
        If `True`, plot the posterior mean and precision on the top of expected mean and
        precision. Defaults to `False`.
    figsize :
        The width and height of the figure. Defaults to `(18, 9)` for a two-level model,
        or to `(18, 12)` for a three-level model.
    color :
        The color of the main curve showing the beliefs trajectory.
    axs :
        A list of Matplotlib axes instances where to draw the trajectories. This should
        correspond to the number of nodes in the structure. The default is `None`
        (create a new figure).

    Returns
    -------
    axs :
        The Matplotlib axes instances where to plot the trajectories.

    Examples
    --------
    Visualization of nodes' trajectories from a three-level continuous HGF model.

    .. plot::

        from pyhgf import load_data
        from pyhgf.model import HGF

        # Set up standard 3-level HGF for continuous inputs
        hgf = HGF(
            n_levels=3,
            model_type="continuous",
            initial_mean={"1": 1.04, "2": 1.0, "3": 1.0},
            initial_precision={"1": 1e4, "2": 1e1, "3": 1e1},
            tonic_volatility={"1": -13.0, "2": -2.0, "3": -2.0},
            tonic_drift={"1": 0.0, "2": 0.0, "3": 0.0},
            volatility_coupling={"1": 1.0, "2": 1.0},
        )

        # Read USD-CHF data
        timeserie = load_data("continuous")

        # Feed input
        hgf.input_data(input_data=timeserie)

        # Plot
        hgf.plot_nodes(node_idxs=1)

    """
    if not isinstance(node_idxs, list):
        node_idxs = [node_idxs]
    trajectories_df = network.to_pandas()

    if axs is None:
        _, axs = plt.subplots(nrows=len(node_idxs), figsize=figsize, sharex=True)

    if isinstance(node_idxs, int) | len(node_idxs) == 1:
        axs = [axs]

    # plotting an input node
    # ----------------------
    for i, node_idx in enumerate(node_idxs):
        if node_idx in network.input_idxs:
            # plotting mean
            axs[i].plot(
                trajectories_df.time,
                trajectories_df[f"x_{node_idx}_expected_mean"],
                label="Expected mean",
                color=color,
                linewidth=1,
                zorder=2,
            )
            axs[i].set_ylabel(rf"$\mu_{{{node_idx}}}$")

            # continuous state node ----------------------------------------------------
            if network.edges[node_idx].node_type == 2:
                input_label = "Continuous"
                axs[i].scatter(
                    x=trajectories_df.time,
                    y=trajectories_df[f"x_{node_idx}_mean"],
                    s=3,
                    label="Input",
                    color="#2a2a2a",
                    zorder=2,
                )
                # plotting standard deviation
                if ci is True:
                    precision = trajectories_df[f"x_{node_idx}_expected_precision"]
                    sd = np.sqrt(1 / precision)
                    y1 = trajectories_df[f"x_{node_idx}_expected_mean"] - sd
                    y2 = trajectories_df[f"x_{node_idx}_expected_mean"] + sd

            # binary state node --------------------------------------------------------
            elif network.edges[node_idx].node_type == 1:
                input_label = "Binary"
                axs[i].scatter(
                    x=trajectories_df.time,
                    y=trajectories_df[f"x_{node_idx}_mean"],
                    label="Input",
                    color="#2a2a2a",
                    zorder=2,
                    alpha=0.4,
                )

                # plotting standard deviation - in the case of a binary input node, the
                # CI should be read from the value parent using the sigmoid transform
                if ci is True:

                    # get parent nodes and sum predictions
                    mean_parent, precision_parent = 0.0, 0.0
                    for idx in network.edges[node_idx].value_parents:  # type: ignore

                        # compute  mu +/- sd at time t-1
                        # and use the sigmoid transform before plotting
                        mean_parent += trajectories_df[f"x_{idx}_expected_mean"]
                        precision_parent += trajectories_df[
                            f"x_{idx}_expected_precision"
                        ]
                    sd = np.sqrt(1 / precision_parent)
                    y1 = 1 / (1 + np.exp(-mean_parent + sd))
                    y2 = 1 / (1 + np.exp(-mean_parent - sd))

            if ci is True:

                axs[i].fill_between(
                    x=trajectories_df["time"],
                    y1=y1,
                    y2=y2,
                    alpha=0.4,
                    color=color,
                    zorder=2,
                )

            axs[i].set_title(f"{input_label} Input Node {node_idx}", loc="left")
            axs[i].legend(loc="upper left")

        # plotting state nodes
        # --------------------
        else:

            axs[i].set_title(
                f"State Node {node_idx}",
                loc="left",
            )

            # show the expected states
            # ------------------------
            # extract sufficient statistics from the data frame
            mean = trajectories_df[f"x_{node_idx}_expected_mean"]

            # plotting mean
            axs[i].plot(
                trajectories_df.time,
                mean,
                label="Expected mean",
                color=color,
                linewidth=1,
                zorder=2,
            )
            axs[i].set_ylabel(rf"$\mu_{{{node_idx}}}$")

            # plotting standard deviation
            if ci is True:
                precision = trajectories_df[f"x_{node_idx}_expected_precision"]
                sd = np.sqrt(1 / precision)
                y1 = mean - sd
                y2 = mean + sd

                axs[i].fill_between(
                    x=trajectories_df.time,
                    y1=y1,
                    y2=y2,
                    alpha=0.4,
                    color=color,
                    zorder=2,
                )

            axs[i].legend(loc="upper left")

            # show the current states
            # -----------------------
            if show_posterior:
                # extract sufficient statistics from the data frame
                mean = trajectories_df[f"x_{node_idx}_mean"]

                axs[i].scatter(
                    x=trajectories_df.time,
                    y=mean,
                    s=3,
                    label="Posterior",
                    color="#2a2a2a",
                    zorder=2,
                    alpha=0.5,
                )
                axs[i].legend(loc="lower left")

        # plotting surprise
        # -----------------
        if show_surprise:
            node_surprise = trajectories_df[f"x_{node_idx}_surprise"].to_numpy()

            if not np.isnan(node_surprise).all():
                surprise_ax = axs[i].twinx()

                sp = node_surprise.sum()
                surprise_ax.set_title(
                    f"Surprise: {sp:.2f}",
                    loc="right",
                )
                surprise_ax.fill_between(
                    x=trajectories_df.time,
                    y1=node_surprise,
                    y2=node_surprise.min(),
                    where=network.node_trajectories[node_idx]["observed"],
                    color="#7f7f7f",
                    alpha=0.1,
                    zorder=-1,
                    label="Surprise",
                )

                # hide surprise if the input was not observed
                node_surprise[network.node_trajectories[node_idx]["observed"] == 0] = (
                    np.nan
                )
                surprise_ax.plot(
                    trajectories_df.time,
                    node_surprise,
                    color="#2a2a2a",
                    linewidth=0.5,
                    zorder=-1,
                )
                surprise_ax.set_ylabel("Surprise")
                surprise_ax.legend(loc="upper right")
    return axs



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

import jax.numpy as jnp
from jax import Array, jit

from pyhgf.typing import Edges


@partial(jit, static_argnames=("node_idx"))
def binary_state_node_prediction_error(
    attributes: Dict, node_idx: int, **args
) -> Array:
    """Compute the value prediction errors and predicted precision of a binary node.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the binary state node.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    """
    # compute the prediction error of the binary state node
    value_prediction_error = (
        attributes[node_idx]["mean"] - attributes[node_idx]["expected_mean"]
    )

    # cancel the prediction error if the value was not observed
    value_prediction_error *= attributes[node_idx]["observed"]

    # scale the prediction error so it can be used in the posterior update
    # (eq. 98, Weber et al., v1)
    value_prediction_error /= attributes[node_idx]["expected_precision"]

    # store the prediction errors in the binary node
    attributes[node_idx]["temp"]["value_prediction_error"] = value_prediction_error

    return attributes


@partial(jit, static_argnames=("edges", "node_idx"))
def binary_finite_state_node_prediction_error(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Update the posterior of a binary node given finite precision of the input.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the node that needs to be updated. After continuous updates, the
        parameters of value and volatility parents (if any) will be different.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.

    Returns
    -------
    attributes :
        The updated attributes of the probabilistic nodes.

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    value_child_idx = edges[node_idx].value_children[0]  # type: ignore

    delata0 = attributes[value_child_idx]["temp"]["value_prediction_error_0"]
    delata1 = attributes[value_child_idx]["temp"]["value_prediction_error_1"]
    expected_precision = attributes[value_child_idx]["expected_precision"]

    # Likelihood under eta1
    und1 = jnp.exp(-expected_precision / 2 * delata1**2)

    # Likelihood under eta0
    und0 = jnp.exp(-expected_precision / 2 * delata0**2)

    # Eq. 39 in Mathys et al. (2014) (i.e., Bayes)
    expected_mean = attributes[node_idx]["expected_mean"]
    mean = expected_mean * und1 / (expected_mean * und1 + (1 - expected_mean) * und0)
    precision = 1 / (expected_mean * (1 - expected_mean))

    attributes[node_idx]["mean"] = mean
    attributes[node_idx]["precision"] = precision

    return attributes



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict, Tuple

import jax.numpy as jnp
from jax import Array, jit, random
from jax._src.typing import Array as KeyArray
from jax.lax import cond
from jax.scipy.stats.norm import pdf
from jax.tree_util import Partial
from jax.typing import ArrayLike

from pyhgf.math import Normal
from pyhgf.typing import Attributes, Edges


@partial(jit, static_argnames=("edges", "node_idx"))
def dirichlet_node_prediction_error(
    edges: Edges,
    attributes: Dict,
    node_idx: int,
    **args,
) -> Attributes:
    """Prediction error and update the child networks of a Dirichlet process node.

    When receiving a new input, this node chose to either:
    1. Allocate the value to a pre-existing cluster.
    2. Create a new cluster.

    The network always contains a temporary branch as the new cluster candidate. This
    branch is parametrized under the new observation to assess its likelihood and the
    previous clusters' likelihood.

    Parameters
    ----------
    edges :
        The edges of the neural network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index lists the value/volatility parents/children.
    attributes :
        The attributes of the probabilistic nodes.
    node_idx :
        Pointer to the Dirichlet process input node.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    """
    values = attributes[node_idx]["mean"]  # the input value
    alpha = attributes[node_idx]["alpha"]  # the concentration parameter
    n_total = attributes[node_idx]["n_total"]  # total number of observations
    n = attributes[node_idx]["n"]  # number of observations per cluster
    sensory_precision = attributes[node_idx][
        "sensory_precision"
    ]  # number of observations per cluster

    # likelihood of the current observation under existing clusters
    # -------------------------------------------------------------
    cluster_ll = clusters_likelihood(
        value=values,
        expected_mean=attributes[node_idx]["expected_means"],
        expected_sigma=attributes[node_idx]["expected_sigmas"],
    )

    # set the likelihood to 0 for inactive clusters
    cluster_ll *= attributes[node_idx]["activated"]

    # likelihood of the current observation under the best candidate cluster
    # ----------------------------------------------------------------------

    # find the best cluster candidate given the new observation
    candidate_mean, candidate_sigma = get_candidate(
        value=values,
        sensory_precision=sensory_precision,
        expected_mean=attributes[node_idx]["expected_means"],
        expected_sigma=attributes[node_idx]["expected_sigmas"],
    )

    # get the likelihood under this candidate
    candidate_ll = clusters_likelihood(
        value=values,
        expected_mean=candidate_mean,
        expected_sigma=candidate_sigma,
    )

    # DP step: compare the likelihood of existing cluster with a new cluster
    # ----------------------------------------------------------------------

    # probability of being assigned to a pre-existing cluster
    cluster_ll *= n / (alpha + n_total)

    # probability to draw a new cluster
    candidate_ll *= alpha / (alpha + n_total)

    best_val = jnp.max(cluster_ll)

    # set all cluster to non-observed by default
    for parent_idx in edges[node_idx].value_parents:  # type:ignore
        attributes[parent_idx]["observed"] = 0

    # get the index of the cluster (!= the node index)
    # depending on whether a new cluster is created or updated
    cluster_idx = jnp.where(
        best_val >= candidate_ll,
        jnp.argmax(cluster_ll),
        attributes[node_idx]["n_active_cluster"],
    )

    update_fn = Partial(
        update_cluster,
        edges=edges,
        node_idx=node_idx,
    )

    create_fn = Partial(
        create_cluster,
        edges=edges,
        node_idx=node_idx,
    )

    # apply either cluster update or cluster creation
    operands = attributes, cluster_idx, values, (candidate_mean, candidate_sigma)

    attributes = cond(best_val >= candidate_ll, update_fn, create_fn, operands)

    attributes[node_idx]["n_total"] += 1

    return attributes


@partial(jit, static_argnames=("edges", "node_idx"))
def update_cluster(operands: Tuple, edges: Edges, node_idx: int) -> Attributes:
    """Update an existing cluster.

    Parameters
    ----------
    operands :
        Non-static parameters.
    edges :
        The edges of the neural network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index lists the value/volatility parents/children.
    node_idx :
        Pointer to the Dirichlet process input node.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    """
    attributes, cluster_idx, value, _ = operands

    # activate the corresponding branch and pass the value
    for i, value_parent_idx in enumerate(edges[node_idx].value_parents):  # type: ignore

        attributes[value_parent_idx]["observed"] = jnp.where(cluster_idx == i, 1.0, 0.0)
        attributes[value_parent_idx]["mean"] = value

    attributes[node_idx]["n"] = (
        attributes[node_idx]["n"]
        .at[cluster_idx]
        .set(attributes[node_idx]["n"][cluster_idx] + 1.0)
    )

    return attributes


@partial(jit, static_argnames=("edges", "node_idx"))
def create_cluster(operands: Tuple, edges: Edges, node_idx: int) -> Attributes:
    """Create a new cluster.

    Parameters
    ----------
    operands :
        Non-static parameters.
    edges :
        The edges of the neural network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as node number.
        For each node, the index lists the value/volatility parents/children.
    node_idx :
        Pointer to the Dirichlet process input node.

    Returns
    -------
    attributes :
        The attributes of the probabilistic nodes.

    """
    attributes, cluster_idx, value, (candidate_mean, candidate_sigma) = operands

    # creating a new cluster
    attributes[node_idx]["activated"] = (
        attributes[node_idx]["activated"].at[cluster_idx].set(1)
    )

    for i, value_parent_idx in enumerate(edges[node_idx].value_parents):  # type: ignore

        attributes[value_parent_idx]["observed"] = 0.0
        attributes[value_parent_idx]["mean"] = value

        # initialize the new cluster using candidate values
        attributes[value_parent_idx]["xis"] = jnp.where(
            cluster_idx == i,
            Normal().expected_sufficient_statistics(
                mu=candidate_mean, sigma=candidate_sigma
            ),
            attributes[value_parent_idx]["xis"],
        )

    attributes[node_idx]["n"] = attributes[node_idx]["n"].at[cluster_idx].set(1.0)
    attributes[node_idx]["n_active_cluster"] += 1

    return attributes


@jit
def get_candidate(
    value: float,
    sensory_precision: float,
    expected_mean: ArrayLike,
    expected_sigma: ArrayLike,
    n_samples: int = 20_000,
) -> Tuple[float, float]:
    """Find the best cluster candidate given previous clusters and an input value.

    Parameters
    ----------
    value :
        The new observation.
    sensory_precision :
        The expected precision of the new observation.
    expected_mean :
        The mean of the existing clusters.
    expected_sigma :
        The standard deviation of the existing clusters.
    n_samples :
        The number of samples that should be simulated.

    Returns
    -------
    mean :
        The mean of the new candidate cluster.
    sigma :
        The standard deviation of the new candidate cluster.

    """
    # sample n likely clusters given the base distribution priors
    mus, sigmas, weights = likely_cluster_proposal(
        mean_mu_G0=0.0,
        sigma_mu_G0=10.0,
        sigma_pi_G0=3.0,
        expected_mean=expected_mean,
        expected_sigma=expected_sigma,
        key=random.key(42),
        n_samples=n_samples,
    )

    # 1 - Likelihood of the new observation under each sampled cluster
    # ----------------------------------------------------------------
    ll_value = pdf(value, mus, sigmas)
    ll_value /= ll_value.sum()  # normalize the weights

    # 2- re-scale the weights using expected precision
    # ------------------------------------------------
    weights *= ll_value**sensory_precision

    # only use the 1000 best candidates for inference
    idxs = jnp.argsort(weights)
    mus, sigmas, weights = (
        mus[idxs][-1000:],
        sigmas[idxs][-1000:],
        weights[idxs][-1000:],
    )

    # 3 - estimate new mean and standard deviation using the weigthed mean
    # --------------------------------------------------------------------
    mean = jnp.average(mus, weights=weights)
    sigma = jnp.average(sigmas, weights=weights)

    return mean, sigma


@partial(jit, static_argnames=("n_samples"))
def likely_cluster_proposal(
    mean_mu_G0: float,
    sigma_mu_G0: float,
    sigma_pi_G0: float,
    expected_mean=ArrayLike,
    expected_sigma=ArrayLike,
    key: KeyArray = random.key(42),
    n_samples: int = 20_000,
) -> Tuple[Array, Array, Array]:
    """Sample likely new belief distributions given pre-existing clusters.

    Parameters
    ----------
    mean_mu_G0 :
        The mean of the mean of the base distribution.
    sigma_mu_G0 :
        The standard deviation of mean of the base distribution.
    sigma_pi_G0 :
        The standard deviation of the standard deviation of the base distribution.
    expected_mean :
        Pre-existing clusters means.
    expected_sigma :
        Pre-existing clusters standard deviation.
    key :
        Random state.
    n_samples :
        The number of samples used during the simulations.

    Returns
    -------
    new_mu :
        A vector of means candidates.
    new_sigma :
        A vector of standard deviation candidates.
    weights :
        Weigths for each cluster candidate under pre-existing cluster (irrespective of
        new observations).

    """
    # sample new candidate for cluster means
    key, use_key = random.split(key)
    new_mu = sigma_mu_G0 * random.normal(use_key, (n_samples,)) + mean_mu_G0

    # sample new candidate for cluster standard deviation
    key, use_key = random.split(key)
    new_sigma = jnp.abs(random.normal(use_key, (n_samples,)) * sigma_pi_G0)

    # 1 - Cluster specificity
    # -----------------------
    # this cluster should explain new dimensions, not explained by other clusters

    # evidence for pre-existing clusters
    pre_existing_likelihood = jnp.zeros(n_samples)
    for mu_i, sigma_i in zip(expected_mean, expected_sigma):
        pre_existing_likelihood += pdf(new_mu, mu_i, sigma_i)

    # evidence for the new cluster proposal
    new_likelihood = pdf(new_mu, new_mu, new_sigma)

    # standardize the measure of cluster specificity (ratio)
    ratio = new_likelihood / (new_likelihood + pre_existing_likelihood)
    ratio -= ratio.min()
    ratio /= ratio.max()
    weights = ratio

    # 2 - Cluster isolation
    # ---------------------
    # this cluster should not try to explain what was already explained

    # (pre-existing cluster) / (pre-existing cluster + new cluster)
    cluster_isolation = jnp.ones(n_samples)
    for mu_i, sigma_i in zip(expected_mean, expected_sigma):
        ratio = pdf(mu_i, mu_i, sigma_i) / (
            pdf(mu_i, mu_i, sigma_i) + pdf(mu_i, new_mu, new_sigma)
        )
        cluster_isolation *= ratio
    cluster_isolation -= cluster_isolation.min()
    cluster_isolation /= cluster_isolation.max()

    weights *= cluster_isolation

    # 3 - Spread of the cluster
    # -------------------------
    # large clusters should be favored over small clusters
    cluster_spread = pdf(1 / (new_sigma**2), 0.0, 5.0)
    cluster_spread -= cluster_spread.min()
    cluster_spread /= cluster_spread.max()
    weights *= cluster_spread

    return new_mu, new_sigma, weights


def clusters_likelihood(
    value: float,
    expected_mean: ArrayLike,
    expected_sigma: ArrayLike,
) -> ArrayLike:
    """Likelihood of a parametrized candidate under the new observation.

    Parameters
    ----------
    value :
        The new observation.
    expected_mean :
        Pre-existing clusters means.
    expected_sigma :
        Pre-existing clusters standard deviation.

    Returns
    -------
    likelihood :
        The probability of observing the value under each cluster.

    """
    return pdf(value, expected_mean, expected_sigma)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax import jit
from jax.tree_util import Partial
from jax.typing import ArrayLike

from pyhgf.math import Normal, binary_surprise, gaussian_surprise
from pyhgf.typing import AdjacencyLists, Attributes, Edges, Sequence, UpdateSequence
from pyhgf.updates.observation import set_observation
from pyhgf.updates.posterior.categorical import categorical_state_update
from pyhgf.updates.posterior.continuous import (
    continuous_node_posterior_update,
    continuous_node_posterior_update_ehgf,
)
from pyhgf.updates.posterior.exponential import posterior_update_exponential_family
from pyhgf.updates.prediction.binary import binary_state_node_prediction
from pyhgf.updates.prediction.continuous import continuous_node_prediction
from pyhgf.updates.prediction.dirichlet import dirichlet_node_prediction
from pyhgf.updates.prediction_error.binary import binary_state_node_prediction_error
from pyhgf.updates.prediction_error.categorical import (
    categorical_state_prediction_error,
)
from pyhgf.updates.prediction_error.continuous import continuous_node_prediction_error
from pyhgf.updates.prediction_error.dirichlet import dirichlet_node_prediction_error
from pyhgf.updates.prediction_error.generic import generic_state_prediction_error

if TYPE_CHECKING:
    from pyhgf.model import Network


@partial(jit, static_argnames=("update_sequence", "edges", "input_idxs"))
def beliefs_propagation(
    attributes: Attributes,
    inputs: Tuple[ArrayLike, ...],
    update_sequence: UpdateSequence,
    edges: Edges,
    input_idxs: Tuple[int],
) -> Tuple[Dict, Dict]:
    """Update the network's parameters after observing new data point(s).

    This function performs the beliefs propagation step. Belief propagation consists in:
    1. A prediction sequence, from the leaves of the graph to the roots.
    2. The assignation of new observations to target nodes (usually the roots of the
    network)
    3. An inference step alternating between prediction errors and posterior updates,
    starting from the roots of the network to the leaves.
    This function returns a tuple of two new `parameter_structure` (i.e. the carryover
    and the accumulated in the context of :py:func:`jax.lax.scan`).

    Parameters
    ----------
    attributes :
        The dictionaries of nodes' parameters. This variable is updated and returned
        after the beliefs propagation step.
    inputs :
        A tuple of n by time steps arrays containing the new observation(s), the time
        steps as well as a boolean mask for observed values. The new observations are a
        tuple of array, with length equal to the number of input nodes. Each input node
        can receive observations  The time steps are the last
        column of the array, the default is unit incrementation.
    update_sequence :
        The sequence of updates that will be applied to the node structure.
    edges :
        Information on the network's edges.
    input_idxs :
        List input indexes.

    Returns
    -------
    attributes, attributes :
        A tuple of parameters structure (carryover and accumulated).

    """
    prediction_steps, update_steps = update_sequence

    # unpack input data - input_values is a tuple of n x time steps arrays
    (*input_data, time_step) = inputs

    attributes[-1]["time_step"] = time_step

    # Prediction sequence
    # -------------------
    for step in prediction_steps:

        node_idx, update_fn = step

        attributes = update_fn(
            attributes=attributes,
            node_idx=node_idx,
            edges=edges,
        )

    # Observations
    # ------------
    for values, observed, node_idx in zip(
        input_data[::2], input_data[1::2], input_idxs
    ):

        attributes = set_observation(
            attributes=attributes,
            node_idx=node_idx,
            values=values,
            observed=observed,
        )

    # Update sequence
    # ---------------
    for step in update_steps:

        node_idx, update_fn = step

        attributes = update_fn(
            attributes=attributes,
            node_idx=node_idx,
            edges=edges,
        )

    return (
        attributes,
        attributes,
    )  # ("carryover", "accumulated")


def list_branches(node_idxs: List, edges: Tuple, branch_list: List = []) -> List:
    """Return the branch of a network from a given set of root nodes.

    This function searches recursively and lists the parents above a given node. If all
    the children of a given parent are on the exclusion list, this parent is also
    excluded.

    Parameters
    ----------
    node_idxs :
        A list of node indexes. The nodes can be input nodes or any other node in the
        network.
    edges :
        The nodes structure.
    branch_list :
        The list of nodes that are already excluded (i.e )

    Returns
    -------
    branch_list :
        The list of node indexes that belong to the branch.

    """
    for idx in node_idxs:
        # add this node to the exclusion list
        branch_list.append(idx)
        all_parents = np.array(
            [
                i
                for i in [
                    edges[idx].value_parents,
                    edges[idx].volatility_parents,
                ]
                if i is not None
            ]
        ).flatten()
        for parent_idx in all_parents:
            # list the children for this node
            all_children = np.array(
                [
                    i
                    for i in [
                        edges[parent_idx].value_children,
                        edges[parent_idx].volatility_children,
                    ]
                    if i is not None
                ]
            ).flatten()
            # if this parent has only excluded children, add it to the exclusion list
            if np.all([i in branch_list for i in all_children]):
                branch_list = list_branches(
                    [parent_idx], edges, branch_list=branch_list
                )

    return branch_list


def fill_categorical_state_node(
    network: "Network",
    node_idx: int,
    binary_states_idxs: List[int],
    binary_parameters: Dict,
) -> "Network":
    """Generate a binary network implied by categorical state(-transition) nodes.

    Parameters
    ----------
    network :
        Instance of a Network.
    node_idx :
        Index to the categorical state node.
    binary_states_idxs :
        The indexes of the binary state nodes.
    binary_parameters :
        Parameters for the set of implied binary HGFs.

    Returns
    -------
    hgf :
        The updated instance of the HGF model.

    """
    # add the binary states - one for each category
    network.add_nodes(
        kind="binary-state",
        n_nodes=len(binary_states_idxs),
        node_parameters={
            "mean": binary_parameters["mean_1"],
            "precision": binary_parameters["precision_1"],
        },
    )

    # add the value coupling between the categorical and binary states
    edges_as_list: List[AdjacencyLists] = list(network.edges)
    edges_as_list[node_idx] = AdjacencyLists(
        5, tuple(binary_states_idxs), None, None, None, (None,)
    )
    for binary_idx in binary_states_idxs:
        edges_as_list[binary_idx] = AdjacencyLists(
            1, None, None, (node_idx,), None, (None,)
        )
    network.edges = tuple(edges_as_list)

    # add continuous state parent nodes
    n_nodes = len(network.edges)
    for i in range(binary_parameters["n_categories"]):
        network.add_nodes(
            value_children=i + n_nodes - binary_parameters["n_categories"],
            node_parameters={
                "mean": binary_parameters["mean_2"],
                "precision": binary_parameters["precision_2"],
                "tonic_volatility": binary_parameters["tonic_volatility_2"],
            },
        )

    # add the higher level volatility parents
    # as a shared parents between the second level nodes
    network.add_nodes(
        volatility_children=[
            idx + binary_parameters["n_categories"] for idx in binary_states_idxs
        ],
        node_parameters={
            "mean": binary_parameters["mean_3"],
            "precision": binary_parameters["precision_3"],
            "tonic_volatility": binary_parameters["tonic_volatility_3"],
        },
    )

    return network


def get_update_sequence(
    network: "Network", update_type: str
) -> Tuple[Sequence, Sequence]:
    """Generate an update sequence from the network's structure.

    This function return an optimized update sequence considering the edges of the
    network. The function ensures that the following principles apply:
    1. all children have computed prediction errors before the parent is updated.
    2. all children have been updated before the parent compute the prediction errors.

    Parameters
    ----------
    network :
        A neural network, instance of :py:class:`pyhgf.model.network.Network`.
    update_type :
        The type of update to perform for volatility coupling. Can be `"eHGF"`
        (defaults) or `"standard"`. The eHGF update step was proposed as an
        alternative to the original definition in that it starts by updating the
        mean and then the precision of the parent node, which generally reduces the
        errors associated with impossible parameter space and improves sampling.

    Returns
    -------
    prediction_sequence :
        The sequence of prediction update.
    update_sequence :
        The sequence of prediction error and posterior updates.

    """
    # initialize the update and prediction sequences
    update_sequence: List = []
    prediction_sequence: List = []

    n_nodes = len(network.edges)

    # list all nodes that are not triggering prediction errors or posterior updates
    # do not call posterior updates for nodes without children (input nodes)
    nodes_without_prediction_error = [i for i in range(n_nodes)]
    nodes_without_prediction = [i for i in range(n_nodes)]
    nodes_without_posterior_update = [
        i
        for i in range(n_nodes)
        if not (
            (network.edges[i].value_children is None)
            & (network.edges[i].volatility_children is None)
        )
    ]

    # prediction updates ---------------------------------------------------------------
    while True:
        no_update = True

        # for all nodes that should apply prediction update ----------------------------
        # verify that all children have computed the prediction error
        for idx in nodes_without_prediction:
            all_parents = [
                i
                for idx in [
                    network.edges[idx].value_parents,
                    network.edges[idx].volatility_parents,
                ]
                if idx is not None
                for i in idx
            ]

            # there is no parent waiting for a prediction update
            if not any([i in nodes_without_prediction for i in all_parents]):
                no_update = False
                nodes_without_prediction.remove(idx)
                if network.edges[idx].node_type == 1:
                    prediction_sequence.append((idx, binary_state_node_prediction))
                elif network.edges[idx].node_type == 2:
                    prediction_sequence.append((idx, continuous_node_prediction))
                elif network.edges[idx].node_type == 4:
                    prediction_sequence.append((idx, dirichlet_node_prediction))

        if not nodes_without_prediction:
            break

        if no_update:
            raise Warning(
                "The structure of the network cannot be updated consistently."
            )

    # prediction errors and posterior updates
    # will fail if the structure of the network does not allow a consistent update order
    # ----------------------------------------------------------------------------------
    while True:
        no_update = True

        # for all nodes that should apply posterior update -----------------------------
        # verify that all children have computed the prediction error
        update_fn = None
        for idx in nodes_without_posterior_update:
            all_children = [
                i
                for idx in [
                    network.edges[idx].value_children,
                    network.edges[idx].volatility_children,
                ]
                if idx is not None
                for i in idx
            ]

            # all the children have computed prediction errors
            if all([i not in nodes_without_prediction_error for i in all_children]):
                no_update = False
                if network.edges[idx].node_type == 2:
                    if update_type == "eHGF":
                        if network.edges[idx].volatility_children is not None:
                            update_fn = continuous_node_posterior_update_ehgf
                        else:
                            update_fn = continuous_node_posterior_update
                    elif update_type == "standard":
                        update_fn = continuous_node_posterior_update

                elif network.edges[idx].node_type == 3:

                    # create the sufficient statistic function
                    # for the exponential family node
                    ef_update = Partial(
                        posterior_update_exponential_family,
                        sufficient_stats_fn=Normal().sufficient_statistics,
                    )
                    update_fn = ef_update

                elif network.edges[idx].node_type == 4:

                    update_fn = None

                update_sequence.append((idx, update_fn))
                nodes_without_posterior_update.remove(idx)

        # for all nodes that should apply prediction error------------------------------
        # verify that all children have been updated
        update_fn = None
        for idx in nodes_without_prediction_error:

            all_parents = [
                i
                for idx in [
                    network.edges[idx].value_parents,
                    network.edges[idx].volatility_parents,
                ]
                if idx is not None
                for i in idx
            ]

            # if this node has no parent, no need to compute prediction errors
            if len(all_parents) == 0:
                nodes_without_prediction_error.remove(idx)
            else:
                # if this node has been updated
                if idx not in nodes_without_posterior_update:

                    if network.edges[idx].node_type == 0:
                        update_fn = generic_state_prediction_error
                    elif network.edges[idx].node_type == 1:
                        update_fn = binary_state_node_prediction_error
                    elif network.edges[idx].node_type == 2:
                        update_fn = continuous_node_prediction_error
                    elif network.edges[idx].node_type == 3:
                        update_fn = None
                    elif network.edges[idx].node_type == 4:
                        update_fn = dirichlet_node_prediction_error
                    elif network.edges[idx].node_type == 5:
                        update_fn = categorical_state_prediction_error

                        # add the update here, this will move at the end of the sequence
                        update_sequence.append((idx, categorical_state_update))
                    else:
                        raise ValueError(f"Invalid node type encountered at node {idx}")

                    no_update = False
                    update_sequence.append((idx, update_fn))
                    nodes_without_prediction_error.remove(idx)

        if (not nodes_without_prediction_error) and (
            not nodes_without_posterior_update
        ):
            break

        if no_update:
            raise Warning(
                "The structure of the network cannot be updated consistently."
            )

    # remove None steps and return the update sequence
    prediction_sequence = [
        update for update in prediction_sequence if update[1] is not None
    ]
    update_sequence = [update for update in update_sequence if update[1] is not None]

    # move all categorical steps at the end of the sequence
    for step in update_sequence:
        if not isinstance(step[1], Partial):
            if step[1].__name__ == "categorical_state_update":
                update_sequence.remove(step)
                update_sequence.append(step)

    return tuple(prediction_sequence), tuple(update_sequence)


def to_pandas(network: "Network") -> pd.DataFrame:
    """Export the nodes trajectories and surprise as a Pandas data frame.

    Returns
    -------
    trajectories_df :
        Pandas data frame with the time series of sufficient statistics and the
        surprise of each node in the structure.

    """
    n_nodes = len(network.edges)
    # get time and time steps from the first input node
    trajectories_df = pd.DataFrame(
        {
            "time_steps": network.node_trajectories[-1]["time_step"],
            "time": jnp.cumsum(network.node_trajectories[-1]["time_step"]),
        }
    )

    # loop over continuous and binary state nodes and store sufficient statistics
    # ---------------------------------------------------------------------------
    states_indexes = [i for i in range(n_nodes) if network.edges[i].node_type in [1, 2]]
    df = pd.DataFrame(
        dict(
            [
                (f"x_{i}_{var}", network.node_trajectories[i][var])
                for i in states_indexes
                for var in network.node_trajectories[i].keys()
                if (("mean" in var) or ("precision" in var))
            ]
        )
    )
    trajectories_df = pd.concat([trajectories_df, df], axis=1)

    # loop over exponential family state nodes and store sufficient statistics
    # ------------------------------------------------------------------------
    ef_indexes = [i for i in range(n_nodes) if network.edges[i].node_type == 3]
    for i in ef_indexes:
        for var in ["nus", "xis", "mean"]:
            if network.node_trajectories[i][var].ndim == 1:
                trajectories_df = pd.concat(
                    [
                        trajectories_df,
                        pd.DataFrame(
                            dict([(f"x_{i}_{var}", network.node_trajectories[i][var])])
                        ),
                    ],
                    axis=1,
                )
            else:
                for ii in range(network.node_trajectories[i][var].shape[1]):
                    trajectories_df = pd.concat(
                        [
                            trajectories_df,
                            pd.DataFrame(
                                dict(
                                    [
                                        (
                                            f"x_{i}_{var}_{ii}",
                                            network.node_trajectories[i][var][:, ii],
                                        )
                                    ]
                                )
                            ),
                        ],
                        axis=1,
                    )

    # add surprise from binary state nodes
    binary_indexes = [i for i in range(n_nodes) if network.edges[i].node_type == 1]
    for bin_idx in binary_indexes:
        surprise = binary_surprise(
            x=network.node_trajectories[bin_idx]["mean"],
            expected_mean=network.node_trajectories[bin_idx]["expected_mean"],
        )
        trajectories_df[f"x_{bin_idx}_surprise"] = surprise

    # add surprise from continuous state nodes
    continuous_indexes = [i for i in range(n_nodes) if network.edges[i].node_type == 2]
    for con_idx in continuous_indexes:
        surprise = gaussian_surprise(
            x=network.node_trajectories[con_idx]["mean"],
            expected_mean=network.node_trajectories[con_idx]["expected_mean"],
            expected_precision=network.node_trajectories[con_idx]["expected_precision"],
        )
        trajectories_df[f"x_{con_idx}_surprise"] = surprise

    # compute the global surprise over all node
    trajectories_df["total_surprise"] = trajectories_df.iloc[
        :, trajectories_df.columns.str.contains("_surprise")
    ].sum(axis=1, min_count=1)

    return trajectories_df


def add_edges(
    attributes: Dict,
    edges: Edges,
    kind="value",
    parent_idxs=Union[int, List[int]],
    children_idxs=Union[int, List[int]],
    coupling_strengths: Union[float, List[float], Tuple[float]] = 1.0,
    coupling_fn: Tuple[Optional[Callable], ...] = (None,),
) -> Tuple:
    """Add a value or volatility coupling link between a set of nodes.

    Parameters
    ----------
    attributes :
        Attributes of the neural network.
    edges :
        Edges of the neural network.
    kind :
        The kind of coupling can be `"value"` or `"volatility"`.
    parent_idxs :
        The index(es) of the parent node(s).
    children_idxs :
        The index(es) of the children node(s).
    coupling_strengths :
        The coupling strength between the parents and children.
    coupling_fn :
        Coupling function(s) between the current node and its value children.
        It has to be provided as a tuple. If multiple value children are specified,
        the coupling functions must be stated in the same order of the children.
        Note: if a node has multiple parents nodes with different coupling
        functions, a coupling function should be indicated for all the parent nodes.
        If no coupling function is stated, the relationship between nodes is assumed
        linear.

    """
    if kind not in ["value", "volatility"]:
        raise ValueError(
            f"The kind of coupling should be value or volatility, got {kind}"
        )
    if isinstance(children_idxs, int):
        children_idxs = [children_idxs]
    assert isinstance(children_idxs, (list, tuple))

    if isinstance(parent_idxs, int):
        parent_idxs = [parent_idxs]
    assert isinstance(parent_idxs, (list, tuple))

    if isinstance(coupling_strengths, int):
        coupling_strengths = [float(coupling_strengths)]
    if isinstance(coupling_strengths, float):
        coupling_strengths = [coupling_strengths]

    assert isinstance(coupling_strengths, (list, tuple))

    edges_as_list = list(edges)
    # update the parent nodes
    # -----------------------
    for parent_idx in parent_idxs:
        # unpack the parent's edges
        (
            node_type,
            value_parents,
            volatility_parents,
            value_children,
            volatility_children,
            this_coupling_fn,
        ) = edges_as_list[parent_idx]

        if kind == "value":
            if value_children is None:
                value_children = tuple(children_idxs)
                attributes[parent_idx]["value_coupling_children"] = tuple(
                    coupling_strengths
                )
            else:
                value_children = value_children + tuple(children_idxs)
                attributes[parent_idx]["value_coupling_children"] += tuple(
                    coupling_strengths
                )
                this_coupling_fn = this_coupling_fn + coupling_fn
        elif kind == "volatility":
            if volatility_children is None:
                volatility_children = tuple(children_idxs)
                attributes[parent_idx]["volatility_coupling_children"] = tuple(
                    coupling_strengths
                )
            else:
                volatility_children = volatility_children + tuple(children_idxs)
                attributes[parent_idx]["volatility_coupling_children"] += tuple(
                    coupling_strengths
                )

        # save the updated edges back
        edges_as_list[parent_idx] = AdjacencyLists(
            node_type,
            value_parents,
            volatility_parents,
            value_children,
            volatility_children,
            this_coupling_fn,
        )

    # update the children nodes
    # -------------------------
    for children_idx in children_idxs:
        # unpack this node's edges
        (
            node_type,
            value_parents,
            volatility_parents,
            value_children,
            volatility_children,
            coupling_fn,
        ) = edges_as_list[children_idx]

        if kind == "value":
            if value_parents is None:
                value_parents = tuple(parent_idxs)
                attributes[children_idx]["value_coupling_parents"] = tuple(
                    coupling_strengths
                )
            else:
                value_parents = value_parents + tuple(parent_idxs)
                attributes[children_idx]["value_coupling_parents"] += tuple(
                    coupling_strengths
                )
        elif kind == "volatility":
            if volatility_parents is None:
                volatility_parents = tuple(parent_idxs)
                attributes[children_idx]["volatility_coupling_parents"] = tuple(
                    coupling_strengths
                )
            else:
                volatility_parents = volatility_parents + tuple(parent_idxs)
                attributes[children_idx]["volatility_coupling_parents"] += tuple(
                    coupling_strengths
                )

        # save the updated edges back
        edges_as_list[children_idx] = AdjacencyLists(
            node_type,
            value_parents,
            volatility_parents,
            value_children,
            volatility_children,
            coupling_fn,
        )

    # convert the list back to a tuple
    edges = tuple(edges_as_list)

    return attributes, edges


def get_input_idxs(edges: Edges) -> Tuple[int, ...]:
    """List all possible default inputs nodes.

    An input node is a state node without any child.

    Parameters
    ----------
    edges :
        The edges of the probabilistic network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the number of
        nodes. For each node, the index list value/volatility - parents/children.

    """
    return tuple(
        [
            i
            for i in range(len(edges))
            if (
                (edges[i].value_children is None)
                & (edges[i].volatility_children is None)
            )
        ]
    )



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict

import jax.numpy as jnp
from jax import Array, jit

from pyhgf.typing import Edges


@partial(jit, static_argnames=("edges", "node_idx"))
def predict_mean(
    attributes: Dict,
    edges: Edges,
    node_idx: int,
) -> Array:
    r"""Compute the expected mean of a continuous state node.

    The expected mean at time :math:`k` for a state node :math:`a` with optional value
    parent(s) :math:`b` is given by:

    .. math::

        \hat{\mu}_a^{(k)} = \lambda_a \mu_a^{(k-1)} + P_a^{(k)}

    where :math:`P_a^{(k)}` is the drift rate (the total predicted drift of the mean,
    which sums the tonic and - optionally - phasic drifts). The variable
    :math:`lambda_a` represents the state's autoconnection strength, with
    :math:`\lambda_a \in [0, 1]`. When :math:`lambda_a = 1`, the node is performing a
    Gaussian Random Walk using the value :math:` P_a^{(k)}` as total drift rate. When
    :math:`\lambda_a < 1`, the state will revert back to the total mean :math:`M_a`,
    which is given by:

    .. math::
            M_a = \frac{\rho_a + f\left(x_b^{(k)}\right)} {1-\lambda_a},

    If :math:`\lambda_a = 0`, the node is not influenced by its own mean anymore, but
    by the value received by the value parent.

    .. hint::

        By combining one parameter :math:`\lambda_a \in [0, 1]` and the influence of
        value parents, it is possible to implement both Gaussian Random Walks and
        Autoregressive Processes, without requiring specific coupling types.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic network that contains the continuous state
        node.
    edges :
        The edges of the probabilistic network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the number of
        nodes. For each node, the index list value/volatility - parents/children.
    node_idx :
        Index of the node that should be updated.

    Returns
    -------
    expected_mean :
        The new expected mean of the state node.

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    time_step = attributes[-1]["time_step"]

    # List the node's value parents
    value_parents_idxs = edges[node_idx].value_parents

    # Get the drift rate from the node
    driftrate = attributes[node_idx]["tonic_drift"]

    # Look at the (optional) value parents for this node
    # and update the drift rate accordingly
    if value_parents_idxs is not None:
        for value_parent_idx, psi in zip(
            value_parents_idxs,
            attributes[node_idx]["value_coupling_parents"],
        ):
            # look at each value parent
            # and get the coupling function to compute the drift
            child_position = edges[value_parent_idx].value_children.index(node_idx)
            coupling_fn = edges[value_parent_idx].coupling_fn[child_position]
            if coupling_fn is None:
                parent_value = attributes[value_parent_idx]["mean"]
            else:
                parent_value = coupling_fn(attributes[value_parent_idx]["mean"])

            driftrate += psi * parent_value

    # The new expected mean from the previous value
    expected_mean = (
        attributes[node_idx]["autoconnection_strength"] * attributes[node_idx]["mean"]
    ) + (time_step * driftrate)

    return expected_mean


@partial(jit, static_argnames=("edges", "node_idx"))
def predict_precision(attributes: Dict, edges: Edges, node_idx: int) -> Array:
    r"""Compute the expected precision of a continuous state node.

    The expected precision at time :math:`k` for a state node :math:`a` is given by:

    .. math::

        \hat{\pi}_a^{(k)} = \frac{1}{\frac{1}{\pi_a^{(k-1)}} + \Omega_a^{(k)}}

    where :math:`\Omega_a^{(k)}` is the *total predicted volatility*. This term is the
    sum of the tonic (endogenous) and phasic (exogenous) volatility, such as:

    .. math::

        \Omega_a^{(k)} = t^{(k)}
        \exp{ \left( \omega_a + \sum_{j=1}^{N_{vopa}} \kappa_j \mu_a^{(k-1)} \right) }


    with :math:`\kappa_j` the volatility coupling strength with the volatility parent
    :math:`j`.

    The *effective precision* :math:`\gamma_a^{(k)}` is given by:

    .. math::

        \gamma_a^{(k) = \Omega_a^{(k)}} \hat{\pi}_a^{(k)}

    This value is also saved in the node for later use during the update steps.


    Parameters
    ----------
    attributes :
        The attributes of the probabilistic network that contains the continuous state
        node.
    edges :
        The edges of the probabilistic network as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the number of
        nodes. For each node, the index list value/volatility - parents/children.
    node_idx :
        Index of the node that should be updated.

    Returns
    -------
    expected_precision :
        The new expected precision of the value parent.
    effective_precision :
        The effective_precision :math:`\gamma_a^{(k)}`. This value is stored in the
        node for later use in the update steps.

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    time_step = attributes[-1]["time_step"]

    # List the node's volatility parents
    volatility_parents_idxs = edges[node_idx].volatility_parents

    # Get the tonic volatility from the node
    total_volatility = attributes[node_idx]["tonic_volatility"]

    # Look at the (optional) volatility parents and add their value to the tonic
    # volatility to get the total volatility
    if volatility_parents_idxs is not None:
        for volatility_parents_idx, volatility_coupling in zip(
            volatility_parents_idxs,
            attributes[node_idx]["volatility_coupling_parents"],
        ):
            total_volatility += (
                volatility_coupling * attributes[volatility_parents_idx]["mean"]
            )

    # compute the predicted_volatility from the total volatility
    predicted_volatility = time_step * jnp.exp(total_volatility)
    predicted_volatility = jnp.where(
        predicted_volatility > 1e-128, predicted_volatility, jnp.nan
    )

    # Estimate the new expected precision for the node
    expected_precision = 1 / (
        (1 / attributes[node_idx]["precision"]) + predicted_volatility
    )

    # compute the effective precision (γ)
    effective_precision = predicted_volatility * expected_precision

    return expected_precision, effective_precision


@partial(jit, static_argnames=("edges", "node_idx"))
def continuous_node_prediction(
    attributes: Dict, node_idx: int, edges: Edges, **args
) -> Dict:
    """Update the expected mean and expected precision of a continuous node.

    Parameters
    ----------
    attributes :
        The attributes of the probabilistic nodes.
    .. note::
        The parameter structure also incorporates the value and volatility coupling
        strength with children and parents (i.e. `"value_coupling_parents"`,
        `"value_coupling_children"`, `"volatility_coupling_parents"`,
        `"volatility_coupling_children"`).
    node_idx :
        Pointer to the node that will be updated.
    edges :
        The edges of the probabilistic nodes as a tuple of
        :py:class:`pyhgf.typing.Indexes`. The tuple has the same length as the node
        number. For each node, the index lists the value and volatility parents and
        children.

    Returns
    -------
    attributes :
        The updated attributes of the probabilistic nodes.

    See Also
    --------
    update_continuous_input_parents, update_binary_input_parents

    References
    ----------
    .. [1] Weber, L. A., Waade, P. T., Legrand, N., Møller, A. H., Stephan, K. E., &
       Mathys, C. (2023). The generalized Hierarchical Gaussian Filter (Version 1).
       arXiv. https://doi.org/10.48550/ARXIV.2305.10937

    """
    # Get the new expected mean
    expected_mean = predict_mean(attributes, edges, node_idx)

    # Get the new expected precision and predicted volatility (Ω)
    expected_precision, effective_precision = predict_precision(
        attributes, edges, node_idx
    )

    # Update this node's parameters

    # 1. input node without volatility parent
    if (
        (edges[node_idx].value_children is None)
        and (edges[node_idx].volatility_children is None)
        and (edges[node_idx].volatility_parents is None)
    ):
        attributes[node_idx]["expected_precision"] = attributes[node_idx]["precision"]

    # 2. regular continuous state node, or input with volatility parent
    else:
        attributes[node_idx]["expected_precision"] = expected_precision

    attributes[node_idx]["temp"]["effective_precision"] = effective_precision
    attributes[node_idx]["expected_mean"] = expected_mean

    return attributes
